/** 
 *   Command line implementation.
 *
 *   Derived from "Camera Control for STAR I."
 *
 *   Included in libiarcontrol packaged from v 0.1.8 (not in lib.)
 *  
 *   (c) Copyright 2004, 2007 Federico Bareilles <fede@iar.unlp.edu.ar>,
 *   Instituto Argentino de Radio Astronomia (IAR).
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *     
 *   The author does NOT admit liability nor provide warranty for any
 *   of this software. This material is provided "AS-IS" in the hope
 *   that it may be useful for others.
 *
 **/

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <limits.h>

#include <curses.h>
#include <readline/readline.h>
#include <readline/history.h>

#include "my_stdio.h"
#include "my_stdlib.h"
#include "line_conf.h"
#include "cmd_tools.h"


static int done = 0;

static int readline_flag = 1;
static int readline_history_flag = 0;
static int readline_alternate_flag = 0;
int interactive_flag = 1;
static int print_info_flag = 1;

#define CMD_PROMPT "cmd> "

static char *cmd_prompt = CMD_PROMPT;
static char *user_history = NULL;
static	char *program_name = NULL;

typedef struct _fd_calback_list {
	int fd;
	ct_fd_callback_t *callback;
	void *data;
	struct _fd_calback_list *next;
} FD_CLABACK_LIST;

static FD_CLABACK_LIST *fd_calback_list;

#define FD_CBL_FD(c) c->fd
#define FD_CBL_CALLBACK(c) c->callback
#define FD_CBL_DATA(c) c->data
#define FD_CBL_NEXT(c) c->next

#define FD_CBL_CALLBACK_RUN(c) FD_CBL_CALLBACK(c)(FD_CBL_FD(c), FD_CBL_DATA(c))
#define FD_CBL_NEW() (FD_CLABACK_LIST *) malloc(sizeof(FD_CLABACK_LIST))
#define FD_CBL_GET_LAST(c) while(FD_CBL_NEXT(c) != NULL) c = FD_CBL_NEXT(c)

static int rl_fd = -1;



/*
COMMAND __commands[] = {
	{ "cd",       0, 0, com_chdir, "[path]", "Change working directory" },	
	{ "export",   0, 0, com_export, "name=value", "Change or add an environment variable" },	
	{ "help",     0, 0, com_help, "[cmd]", "Display this text" },
	{ "?",        0, 1, com_help, "[cmd]", "Synonym for `help'" },
	{ "quit",     0, 0, com_quit, "", "Quit ..." },
	{ "q",        0, 1, com_quit, "", "Quit :-)" },
//	{ "test",     0, 0, com_test,"", "[!]" },
	{ (char *)NULL, 0, 0, (rl_icpfunc_t *)NULL, (char *) NULL, (char *)NULL }
};
*/

/* static COMMAND *commands = __commands; */
static COMMAND *commands = NULL;

/* Generator function for command completion.  STATE lets us
   know whether to start from scratch; without any state
   (i.e. STATE == 0), then we start at the top of the list. */
static char *command_generator (char *text, int state)
{ 
	static int list_index, len;
	char *name;

	/* If this is a new word to complete, initialize now.  This
	   includes saving the length of TEXT for efficiency, and
	   initializing the index variable to 0. */
	if (!state) {
		list_index = 0;
		len = strlen (text);
	}

	/* Return the next name which partially matches from the
	   command list. */
	while ((name = commands[list_index].name)) {
		list_index++;
		if (strncmp (name, text, len) == 0)
			return (strdup(name));
	}

	/* If no names matched, then return NULL. */
	return ((char *)NULL);
}


static char **fileman_completion (const char *text, int start, int end)
{
	char **matches;

	matches = (char **)NULL;

	/* If this word is at the start of the line, then it is a
	   command to complete.  Otherwise it is the name of a file in
	   the current directory.  */
	if (start == 0)
		matches = rl_completion_matches (text, (rl_compentry_func_t *) command_generator);
	
	return (matches);
}


/* Tell the GNU Readline library how to complete.  We want to try to
   complete on command names if this is the first word in the line, or
   on filenames if not. */
static void initialize_readline (void)
{
	Keymap km;
	/* char f9_macro[]={"\\C-p\\n"}; */ /* last command */ 
//	char f9_macro[]={"pal last\\n"};
//	char f5_macro[]={"expose\\n"};
//	char f6_macro[]={"pal last\\n"};

	/* Allow conditional parsing of the ~/.inputrc file. */
	rl_readline_name = cmd_prompt;
	
	/* Tell the completer that we want a crack first. */
	rl_attempted_completion_function = fileman_completion;

	/* Read keybindings and variable assignments from filename */
	/* int rl_read_init_file (const char *filename); */

	km = rl_get_keymap_by_name("emacs"); 
//	rl_macro_bind("\e[15~", f5_macro, km);
//	rl_macro_bind("\e[17~", f6_macro, km);
//	rl_macro_bind("\e[20~", f9_macro, km);
}


int cmd_tools_un_link_commands(void)
{
	int i;

	for (i = 0; commands[i].name; i++) {
		if ( commands[i].link ) {
			unlink(commands[i].name);
		}
	}

	return 0;
}


int cmd_tools_link_commands_to(char *program_name)
{
	int i;

	for (i = 0; commands[i].name; i++) {
		if ( commands[i].link ) {
			unlink(commands[i].name);
			if (symlink(program_name, commands[i].name) != 0)
				perror(": ");
		}
	}

	return 0;
}


/* Look up NAME as the name of a command, and return a pointer to that
   command.  Return a NULL pointer if NAME isn't a command name. */
static COMMAND *find_command (char *name)
{
	int i;

	for (i = 0; commands[i].name; i++)
		if (strcmp (name, commands[i].name) == 0)
			return (&commands[i]);

	return ((COMMAND *)NULL);
}


/* Execute a command line. */
static int execute_line (char *line)
{
	int i;
	COMMAND *command;
	char *word;
	int pos_null = 0;
	/* Isolate the command word. */
	i = 0;
	while (line[i] && WHITESPACE(line[i]))	i++;
	word = line + i;

	while (line[i] && !WHITESPACE (line[i])) i++;

	if (line[i]){
		pos_null = i;
		line[i++] = '\0';
	}
	command = find_command (word);
	if (!command) {
		if(pos_null)
			line[pos_null] = ' ';
		return my_system(line);
	}

	/* Get argument to command, if any. */
	while (WHITESPACE (line[i])) i++;

	word = line + i;

	/* Call the function. */
	return ((*(command->func)) (word));
}


/* The user wishes to quit using this program.  Just set DONE
   non-zero. */
int com_quit (char *arg)
{
	done = 1;
	return (0);
}


/* Print out help for ARG, or for all of the commands if ARG is not
   present. */
int com_help (char *arg)
{
	int i, j;
	int printed = 0;
	int cmdlen;
	const int param_legend = 22;

	for (i = 0; commands[i].name; i++) {
		if ((!*arg && commands[i].show) || 
		    (strcmp (arg, commands[i].name) == 0) ) {
			cmdlen = strlen(commands[i].name) + 
				strlen(commands[i].param) + 1;
			printf ("%s %s", commands[i].name, 
				commands[i].param);
			if (cmdlen >= param_legend) {
				printf("\n");
				cmdlen = 0;
			}
			for(j = cmdlen; j < param_legend; j++)
				printf(" ");
			printf ("%s.\n", commands[i].doc);

			printed++;
		}
	}
	if (!printed) {
		printf ("No commands match `%s'.  Possibilties are:\n", arg);

		for (i = 0; commands[i].name; i++) {
			/* Print in six columns. */
			if (printed == 6) {
				printed = 0;
				printf ("\n");
			}
			
			printf ("%s\t", commands[i].name);
			printed++;
		}

		if (printed)
			printf ("\n");
	}
	return (0);
}


#define COMMAND_LINE_LEGEND 80


static char *local_readline(char *prompt)
{
        char *line = NULL;
        fd_set rfds;
        struct timeval tv;
        int retval;
        char c;
        int l;

	write(0, prompt, strlen(prompt));
        line = (char *) malloc(COMMAND_LINE_LEGEND); 

        /* Loop reading and executing lines until the user quits. */
        l = 0;
        do {
                FD_ZERO(&rfds);
                FD_SET(0, &rfds);
                tv.tv_sec = 10;
                tv.tv_usec = 0;
                retval = select(1, &rfds, NULL, NULL, &tv);
                c = 0;
                if (retval) {
                        if (FD_ISSET(0, &rfds)) {
                                /* Win telnet clients send char by char ... */
                                if( read(0, &c, 1) != 1) {
                                        free(line);
                                        return NULL;
                                }
                                *(line+l) = c;
                                l++;
                        }
                }
        } while(c != '\n' && l < COMMAND_LINE_LEGEND && !done);
        if ( *(line+l-2) == 0xd ) *(line+l-2) = 0;
        else *(line+l-1) = 0;

        return line;
}


static void local_add_history (char *line)
{
	HIST_ENTRY *history;

	if ( history_length ) {
		history = history_get(history_length);
		if (strcmp(line, history->line) == 0) {
			return;
		}
	}
	if (strncmp(line, "q", 1) == 0) {
		return;
	}

	add_history (line);

	return;
}


static void pre_execute_line(char *line)
{
	char *s = stripwhite (line);
	
	if (*s)	{
		if ( readline_flag )
			local_add_history (s);
		execute_line (s);
		free(line);
	}

	return;
}


static int wrapper_rl_callback_read_char(int fd, void* data)
{
	rl_callback_read_char();
	
	return 0;
}


static int readline_alternate_method(void)
{
	fd_set rfds;
	int max_fd = 0;
        struct timeval tv;
        int retval;
	FD_CLABACK_LIST *clp = fd_calback_list;
	int force_rl_redisplay_flag;

	if ( clp == NULL) {
		PVERB(PERR, "Incorrect use of readline_alternate_method");
		return -1;
	}
	FD_ZERO(&rfds);
	tv.tv_sec = 10;
	tv.tv_usec = 0;
	do {
		FD_SET(FD_CBL_FD(clp), &rfds);
		if (max_fd < FD_CBL_FD(clp)) max_fd = FD_CBL_FD(clp);
		clp = FD_CBL_NEXT(clp);
	} while (clp != NULL);
	retval = select(max_fd + 1, &rfds, NULL, NULL, &tv);
	if (retval) {
		force_rl_redisplay_flag = 0;
		clp = fd_calback_list;
		do {
			if (FD_ISSET(FD_CBL_FD(clp), &rfds)) {
				FD_CBL_CALLBACK_RUN(clp);
				if (FD_CBL_FD(clp) != rl_fd )
					force_rl_redisplay_flag = 1;
			}
			clp = FD_CBL_NEXT(clp);
		} while (clp != NULL);
		if (force_rl_redisplay_flag ) {
			fd_printf(rl_fd,"\r");
			rl_forced_update_display();
		}

	}

	return retval;	
}


static void cmd_tools_fd_callback_remove_all(void)
{
	FD_CLABACK_LIST *clp = fd_calback_list;
	
	while(FD_CBL_NEXT(clp) != NULL) {
		clp = FD_CBL_NEXT(clp);
		free(fd_calback_list);
		fd_calback_list = clp;
	}

	fd_calback_list = NULL;
}


int cmd_tools_add_fd_callback(int fd, ct_fd_callback_t *fd_callback, void *data)
{
	FD_CLABACK_LIST *clp = NULL;

	PVERB(PINFO,"Alternate method");
	readline_alternate_flag = 1;
	if (fd_callback == NULL) {
		PVERB(PERR, "Invalid fd_callback: fd(%d) ingnored", fd);
		return -1;
	}
	if (fd_calback_list == NULL) /* first */
		clp = fd_calback_list = FD_CBL_NEW();
	else {
		clp = fd_calback_list;
		FD_CBL_GET_LAST(clp);
		FD_CBL_NEXT(clp) =  FD_CBL_NEW();
		clp = FD_CBL_NEXT(clp);
	}
	FD_CBL_FD(clp) = fd;
	FD_CBL_CALLBACK(clp) = fd_callback;
	FD_CBL_DATA(clp) = data;
	FD_CBL_NEXT(clp) = NULL;

	return 0;
}




static int command_main(void)
{
	char *line = NULL;
	char *history_file = (char *) malloc(_POSIX_PATH_MAX * sizeof(char));
	int err;

	if ( readline_flag ) {
		initialize_readline ();	/* Bind our completer. */
	}
	if ( readline_history_flag ) {
		sprintf(history_file,"%s/%s", getenv("HOME"), user_history);
		if (((err = read_history(history_file)) != 0) && (err != 2)) {
			PVERB(PERR, 
				"Error [%d] reading READLINE history file.\n",
				err);
		}
	}
	if(readline_alternate_flag) {
		rl_fd = (rl_instream != NULL ? fileno(rl_instream):0);
		/* FILE * rl_instream 
		   The stdio stream from which Readline reads
		   input. If NULL, Readline defaults to stdin.
		*/
		cmd_tools_add_fd_callback(rl_fd, wrapper_rl_callback_read_char,
					  NULL);
		rl_callback_handler_install(cmd_prompt, pre_execute_line);
	}
	/* Loop reading and executing lines until the user quits. */
	for ( ; done == 0; ) {
		if ( readline_flag ){
			if(readline_alternate_flag) {
				readline_alternate_method();
			} else
				line = readline (cmd_prompt);
		} else 
                        line = local_readline (cmd_prompt);
		if ( ! readline_alternate_flag ) {
			if (!line)
				break;
			/* line is freed in pre_execute_line() */
			pre_execute_line(line);
			/* free (line); */ 
		}
	}
	if ( readline_history_flag ) {
		if (((err = write_history(history_file)) != 0) && (err != 2)) {
			PVERB(PERR, 
			      "Error [%d] writing READLINE history file.\n",
			      err);
		}
	}
	free(history_file);
	if ( readline_alternate_flag) {
		cmd_tools_fd_callback_remove_all();
		rl_callback_handler_remove();
	}

	return (0);
}


int com_chdir (char *arg)
{
	char *path;
	int ret;

	if (*arg == 0)
		path = getenv("HOME");
	else
		path = arg;
	ret = chdir(path);
	if( ret != 0)
		perror("cd");

	return ret;
}


int com_export (char *arg)
{
	int ret;
	char *name, *value, *comment;

	if ((ret = parse_line(arg, &name, &value, &comment)) == 0) {
		/* don't use putenv() hear */
		if ((ret = setenv(name, value,1)) != 0)
			perror("export");
	}
	return ret;
}


static void cartel(void)
{
	static const char *msg[] =
	{
	"",
	"CMD - Engine interface",
	"",
	"(c) Copyright 2007 Federico Bareilles <fede@iar.unlp.edu.ar>,",
	"",
	"This program is free software; you can redistribute it and/or",
	"modify it under the terms of the GNU General Public License as",
	"published by the Free Software Foundation; either version 2 of",
	"the License, or (at your option) any later version.",
	"",     
	"The author does NOT admit liability nor provide warranty for any",
	"of this software. This material is provided \"AS-IS\" in the hope",
	"that it may be useful for others.",
	""
	};
	unsigned j;
	for (j = 0; j < sizeof(msg)/sizeof(msg[0]); j++)
		fprintf(stderr, "\t%s\n", msg[j]);
	
	return;
}


int cmd_tools_set_command(COMMAND *cmd)
{
	int ret = -1;

	if ( cmd != NULL) {
		commands = cmd;
		ret = 0;
	}

	return ret;
}


int cmd_tools_set_prompt(char *prompt)
{
	int ret = -1;

	if ( prompt != NULL) {
		cmd_prompt = prompt;
		ret = 0;
	}

	return ret;
}


int cmd_tools_use_history(char *basename)
{
	int ret = -1;
	int len;

	if ( basename != NULL) {
		readline_history_flag = 1;
		len = strlen (basename) + 10;
		user_history = (char *) malloc(len * sizeof(char));
		snprintf(user_history, len, ".%s_history", basename);
		ret = 0;
	}

	return ret;
}


int cmd_tools_use_baner(int flag)
{
	if (flag ) print_info_flag = 1;
	else print_info_flag = 0;

	return 0;
}


int cmd_tools_main_loop(int argc, char * const argv[])
{
	int ret = 0;
	COMMAND *command;
	char line[_POSIX_ARG_MAX];
	char *cp;

	/* who is my name? */
        if ((cp = strrchr(argv[0], '/')) != NULL) program_name  = cp + 1;
        else program_name  = argv[0];

	command = find_command(program_name);
	if (!command && argv[optind]) {
		command = find_command(argv[optind]);
		if (command) optind++;
	}
	if (command) {
		readline_flag = 0;
		readline_history_flag = 0;
		interactive_flag = 0;
		line[0] = 0;
		if (optind < argc) {
			while (optind < argc) {
				strcat(line, argv[optind++]);
				if (optind < argc)
					strcat(line, " ");
			}
		}
		ret = ((*(command->func)) (line));
	} else if(ret == 0) {
		if (print_info_flag) cartel();
		command_main();
	}
	fd_printf(rl_fd,"\r");
	
        return ret;
}

